# REBOTNIX Main Demo Application
# 21.07.2022
# G. Hilgemann & L. Ickler for REBOTNIX GmbH, GERMANY
# All rights resvered

import sys
import os

#check for root_path variable
try:
  sys.path.append(os.environ['ROOT_PATH'])
except:
  print("Please set the variable ROOT_PATH.")
  sys.exit(0)

import math
import random
import os
import cv2
import numpy as np
import time
import libs.rb_visiontools as rb_visiontools
import sys
import datetime
import pytz
import codecs
import json
import subprocess

#uncomment this line if run manually with alliedvision
#os.environ["GENICAM_GENTL64_PATH"] = ":/opt/pylon/lib/gentlproducer/gtl:/opt/ssd/install_vimba/VimbaGigETL/CTI/arm_64bit:/opt/ssd/install_vimba/VimbaUSBTL/CTI/arm_64bit"

# Application version
version = "1.5.1"
print("RB-Visiontools for GUSTAV, Version:",version)

# config data
cfg_data = None

# setup a logger
import libs.rb_logger as log
day = datetime.date.today().strftime("%d_%m_%Y")
if "ROOT_PATH" in os.environ:
  logdir = os.path.join(os.environ["ROOT_PATH"],"logs")
else:
  logdir = "./logs"
os.makedirs(logdir, exist_ok=True)
logfile = os.path.join(logdir ,"main_" + day + ".log")

# set logger name
l = log.rb_logger().createlogger("main",logfile)

#REBOTNIX Camera SDK
import libs.rb_camera_basler as rb_c
#import libs.rb_camera_alliedvision as rb_c

#get a vailable cameras
available_cameras = rb_c.listCameras()
print(available_cameras)

#if no camera available exit program
if len(available_cameras)==0:
  print("Sorry. No camera is available. Exiting now.")
  l.info("Sorry. No camera is available. Exiting now.")
  sys.exit()

# init rebotnix camera, needs to be initalized with a serial - here first in list
rb_cam = rb_c.rb_camera(available_cameras[0]["serial"], usbreset=True)

# REBOTNIX GPS Module
from libs.rb_gps import rb_gps
#initialize gps
mygps = rb_gps()

# this is required for our neuronal network
netMain = None
metaMain = None
altNames = None

# compression ratio for all stored JPG´s
jpgcomratio = 85

# rebotnix filter class update
from rb_filter_upload import filter_upload_rebotnix

#time class
def is_time_between(begin_time, end_time, check_time=None):
    # If check time is not given, default to current UTC time
    check_time = check_time or datetime.datetime.utcnow().time()
    if begin_time < end_time:
        return check_time >= begin_time and check_time <= end_time
    else: # crosses midnight
        return check_time >= begin_time or check_time <= end_time


# rebotnix detection class
def rb_detect(_treshold, savepath, mode="crop"):

    global rb_cam
    global metaMain, netMain, altNames

    # path to REBOTNIX *.rvt Modelfile
    weights = "../models/main_traffic_model.rvt"
    #check if weight path exists
    if not os.path.exists(weights):
      print("Sorry. Your model file was not found. Exiting application now.")
      l.info("Sorry. Your model file was not found. Exiting application now.")
      rb_cam.stop()
      mygps.stop()
      sys.exit()

    # load a model
    network, class_names, class_colors, width, height = rb_visiontools.load_model_rb(weights)

    # checkexittime
    checkexittime(savepath, mode)

    # calculate start_time
    start_time = time.time()
    check_time = time.time()
    last_gps = -1
    gps_time = time.time()

    # start detection loop
    while True:
        try:
            #check time for upload
            r = checkexittime(savepath, mode)

            # when we have exit time, we disconnect
            if r == True:
                break

            # prev_time
            prev_time = time.time()

            # read current frame
            flag, frame_read = rb_cam.getFlagAndFrame()

            # check if valid image + valid gps
            if flag and mygps.flag:
                start_time = time.time()

                #check if new frame is valid - if not continue
                if np.shape(frame_read) == ():
                  l.info("Empty frame")
                  continue

                # get now time
                nowt = time.time()

                #run a detection
                detections, scale_x, scale_y = rb_visiontools.do_detection_rb(network, class_names, frame_read, width, height, _treshold)
                print(detections)

                # we return GPS based lon, lat, accuracy and number of stats in sat variable
                lon, lat, gps_acc, sat, speed, headAcc, headMot = mygps.getgps()
                #print(lon, lat, gps_acc, sat, speed, headAcc, headMot)

                #break loop if gps not updating after 1 minute
                if last_gps != lon:
                  last_gps = lon
                  gps_time = time.time()
                if gps_time < time.time()-60:
                  mygps.flag = False

                # save an image every minute - as camera sanity check
                if time.time()-check_time>60:
                  check_time = time.time()
                  cv2.imwrite("./testing/"+str(check_time)+"_"+str(lat)+"_"+str(lon)+".jpg",frame_read)

                # do we have an AI detections from model?
                if len(detections) > 0:

                    # train mode stores all jpg
                    if mode == "train":
                        cv2.imwrite(os.path.join("./images_train",str(nowt) + "_" + str(lat) + "_" + str(lon) + ".jpg"),frame_read,[int(cv2.IMWRITE_JPEG_QUALITY),jpgcomratio])

                    # full mode is an internal remondis test case, not in production in the offical product
                    if mode == "full":
                        path = os.path.join("./images_full/", str(nowt) + "_" + str(lat) + "_" + str(lon) + "_full.jpg")
                        # exposetimetxt
                        exposetimetxt = "Expose_low:" + str(rb_cam.getExposeLowerTime()) + " Expose_high:" + str(rb_cam.getExposeUpperTime()) + " Current Expose:" + str(rb_cam.getExposeTime())
                        # give camera information to picture
                        cv2.putText(frame_read,str(exposetimetxt),(25, 55),cv2.FONT_HERSHEY_SIMPLEX,0.8,[255,255,255],2)

                # iterate detections from AI model
                for ind, d in enumerate(detections):

                    # get label
                    label = str(d[0])

                    # get accuracy
                    acc = d[1]

                    # convert yolo coordinates back to readable pixel values
                    x = d[2][0] * scale_x
                    y = d[2][1] * scale_y
                    w = d[2][2] * scale_x
                    h = d[2][3] * scale_y
                    x1,y1,x2,y2 = rb_visiontools.convertBack_rb(float(x), float(y), float(w), float(h))

                    print("Acc: "+str(float(acc)))
                    # check if accuracy above threshold
                    if (float(acc) > float(_treshold)):

                        # debug out / good GPS values are <3, bad signals starts at 5 till >100
                        if mode == "crop" and gps_acc < 4:

                            # select the roi crop of the image
                            crop = frame_read[y1:y2, x1:x2]

                            # shape limitations
                            sh = crop.shape[0]
                            sw = crop.shape[1]

                            # limit shapes width and height - min size
                            if sw > 35 and sh > 35 and sw < 450 and sh < 450:

                                # mesuare darklevel
                                darklevel = np.mean(crop)

                                # we store only bright images
                                if darklevel>5:
                                    print("save")
                                    print(crop.shape)
                                    # write image to disk
                                    cv2.imwrite(os.path.join(savepath,label+"_"+str(nowt) + "_" + str(lat) + "_" + str(lon) + "_" + str(acc) + ".jpg"),crop,[int(cv2.IMWRITE_JPEG_QUALITY),jpgcomratio])
                                else:
                                    print("Image is too dark, we do not save this!")
                                    l.info("Image is too dark, we do not save this!")
                            else:
                                print("Image size is not valid. Crop Shape was:" + str(crop.shape))
                                l.info("Image size not valid. Crop Shape was:" + str(crop.shape))

                        # gps acc
                        if gps_acc > 4:
                            l.info("Gps quality is bad: " + str(gps_acc) + " satellites: " + str(sat))

                        # if mode is full
                        if mode == "full":
                            start_point = (x1, y1) 
                            end_point = (x2, y2)
                            # draw detection rectangle
                            cv2.rectangle(frame_read, start_point, end_point,(0, 255, 0),2)

                        # debug only, print out the fps output
                        #print("FPS: " + str(1/(time.time()-prev_time)))

                    else:
                        print("No saved label:" + label + " Acc is too low: " + str(acc))

                    if mode == "full":
                        cv2.imwrite(path,frame_read,[int(cv2.IMWRITE_JPEG_QUALITY),jpgcomratio])

            else:
                #check if no valid flags for over 30 seconds -> exit application.
                if time.time()-start_time > 30:
                  l.error("No valid flags, we exit and wait for the next start!")
                  print("No valid flags, we exit and wait for the next start!")
                  rb_cam.stop()
                  mygps.stop()
                  sys.exit()

        except Exception as e:
            rb_cam.stop()
            mygps.stop()
            l.error("Something is going wrong.We restart now! Error: "+ str(e))
            print("Something is going wrong.We restart now! Error: "+ str(e))
            sys.exit()

# check exit time
def checkexittime(savepath, mode):
    global cfg_data

    now=datetime.datetime.now()
    zeit = int(now.strftime('%H%M'))

    #get upload times from config
    starttime = datetime.datetime.strptime(cfg_data["uploadparameter"]["starttime"], '%Y-%m-%dT%H:%M:%S.%fZ')
    endtime = datetime.datetime.strptime(cfg_data["uploadparameter"]["endtime"], '%Y-%m-%dT%H:%M:%S.%fZ')
    its_time = is_time_between(datetime.time(starttime.hour,starttime.minute), datetime.time(endtime.hour,endtime.minute), check_time=datetime.time(now.hour, now.minute))

    # it´s time to exit
    if its_time:
        #if mode is full we exit, skip data upload
        if mode == "full":
            rb_cam.stop()
            mygps.stop()
            l.info("It is time to exit the application.")
            print("It is time to exit the application.")
            sys.exit()

        print("we stop GPS and camera now!")
        rb_cam.stop()
        mygps.stop()

        print("Uploading startet",zeit)
        l.info("Data filtering starts")
        rb = filter_upload_rebotnix(savepath, debug=debug_mode)
        rb.filter()
        l.info("Data filtering complete")
        l.info("Uploading starting")
        l.info("Clean up")
        rb.upload()
        l.info("Uploading done")
        l.info("Exiting.")
        os._exit(1)

def main(args=None):

    global carid
    global debug_mode
    global cfg_data

    #read config file 
    if "ROOT_PATH" in os.environ:
      cfg = open(os.path.join(os.environ["ROOT_PATH"],"cfg","main.json"), "r")
    else:
      cfg = open("./config.json", "r")
    text_cfg = cfg.read()
    decoded_cfg=codecs.decode(text_cfg.encode(), 'utf-8-sig')
    cfg_data = json.loads(decoded_cfg)

    #set parameters
    debug_mode = cfg_data["debug_mode"]
    carid = cfg_data["carid"]
    savepath = cfg_data["save_path"]
    mode = cfg_data["imagemode"]

    if not os.path.exists(savepath):
        os.makedirs(savepath)
    if not os.path.exists("./images_full"):
        os.makedirs("./images_full")
    if not os.path.exists("./images_train"):
        os.makedirs("./images_train")
    if not os.path.exists("./testing"):
        os.makedirs("./testing")

    checkexittime(savepath, mode)

    # start detection loop
    rb_detect(float(cfg_data["accuracy"]), savepath, mode)

if __name__ == "__main__":
    main()

